(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open Def;;
open Term;;
open Varset;;
open Varmap;;
open Rewrite;;
open Alpha;;
open Unfold;;
open Printf;;
open Pprinter;;

(*************************)

let rec nomatch t bv =
  match t with
    | Obj o -> true
    | Cste x -> true
    | Var x -> true
    | Type -> true
    | Lambda (x, t1,t2) -> (nomatch  t1 (x ++ bv)) && (nomatch  t2 (x ++ bv))
    | Forall (x, t1,t2) -> (nomatch  t1 (x ++ bv)) && (nomatch  t2 (x ++ bv))
    | Ind (x, t1,t2, t3) -> true
    | Phi (x, t1,t2,n, t3) -> (nomatch  t2 (x ++ bv)) && 
          (nomatch  t3 (x ++ (bv +++ (listvarterm2varset t1))))
    | Cons (n, t1) -> (nomatch  t1 bv)
    | App (t1,t2) -> (nomatch t1 bv) && (nomatch_list t2 bv)
    | Match (t1, t2, t3) -> (
        match t1 with
          | Var x -> if (VarSet.subset (VarSet.singleton x) bv) then (nomatch_list t2 bv) else false
          | _ -> false
      )
and nomatch_list l bv =
  match l with
    | [] -> true
    | hd::tl -> (nomatch hd bv) && (nomatch_list tl bv);;

let rec simpl t def =
  (*
    printf "simpl_term: ";
    let l = vmext def in
    
    printf "Definitions:\n";
    
    List.fold_left (
    
    fun tl hd ->
        let (name, (te, ty)) = hd in
    printf "%s := %s : %s\n"
    name
    (string_of_term te VarMap.empty)
    (string_of_term ty VarMap.empty);
    
    ) () l; printf "\n";
    
    
    print_term t VarMap.empty;
    printf "\n\n";
  *)
  match (app_nf t) with
    | App (t1 , t2) -> (
        match t1 with
          | Lambda (x, t11, t12) -> (
              match t2 with
                | [] -> t1
                | hd :: [] ->
                    let t12' = (rewrite_term_var_term (alpha_term_vars t12 (term_fv hd VarSet.empty)) x hd) in
                      simpl t12' def
                | hd :: tl ->
                    let t12' = (rewrite_term_var_term (alpha_term_vars t12 (term_fv hd VarSet.empty)) x hd) in
                      simpl (App (t12', tl)) def
            )
          | Phi (x, largs, ty, ldecargs, te) -> (
              if (List.length t2 < List.length largs) then
                t
              else (
                let te' = (rewrite_term_var_term te x (Phi (x, largs, ty, ldecargs, te))) in
                let t1' = (largs_lambdas te' largs) in
                let t' = simpl (App (t1', t2)) def in
                  if (nomatch t' VarSet.empty) then
                    t' 
                  else
                    t
              )
            )
          | App (t11, t12) -> (
              simpl (app_nf t) def
            )
          | ( Cste x ) -> (
              let t1' = (app_nf (unfold_term t1 def)) in
                match t1' with
                  | Lambda (_,_,_) -> 
                      if (List.length t2 >= 1) then
                        simpl (App (t1', t2)) def
                      else
                        t
                  | Phi (v, largs, ty, ldecargs, te) -> (
                      if (List.length t2 < List.length largs) then
                        t
                      else (
                        let te' = (rewrite_term_var_term te v (Cste x)) in
                        let t1' = (largs_lambdas te' largs) in
                        let t' = simpl (App (t1', t2)) def in
                          if (nomatch t' VarSet.empty) then
                            t' 
                          else (
                            t
                          )
                      )
                    )
                  | Cste c -> simpl (App (t1', t2)) def
                  | Obj o -> simpl (App (t1', t2)) def
                  | App (t11, t12) -> (                      
                      match t11 with 
                          (*
                        | Lambda (_,_,_) -> 
                            if (List.length (t12 @ t2) >= 1) then
                              simpl (App (t11, (t12 @ t2))) def
                            else
                              t
                        | Phi (v, largs, ty, ldecargs, te) -> (
                            if (List.length (t12 @ t2) < List.length largs) then
                              t
                            else (
                              let te' = (rewrite_term_var_term te v (Phi (v, largs, ty, ldecargs, te))) in
                              let t1' = (largs_lambdas te' largs) in
                              let t' = simpl (App (t1', t12 @ t2)) def in
                                if (nomatch t' VarSet.empty) then
                                  t' 
                                else (
                                  t
                                )
                            )
                          )
                          *)
                        | Cste c -> simpl (App (Cste c, t12 @ t2)) def
                        | _ -> App (t1, simpl_listterm t2 def)
                    )
                  | _ -> App (t1, simpl_listterm t2 def)
            )
          | Match (t11,ldes, t12) ->
              let t1' = simpl (Match (t11, ldes, t12)) def in
                if (t1 = t1') then
                  (app_nf t)
                else
                  simpl (App (t1', t2)) def                
          | Obj o -> (
              o#exec (simpl_listterm t2 def)
            )
          | _ -> App (t1, t2)
      )
    | Match (t11,ldes, t12) ->
        let t11' = simpl t11 def in
        let (t11, t12') = (decomp_app t11') in (            
          match t11 with
            | (Cons (n, _)) ->
                simpl (App (List.nth ldes n, t12')) def
            | Cste c ->
                simpl (Match (App (unfold_term (Cste c) def, t12'), ldes, t12)) def
            | _ -> Match (t11', ldes, t12)
        )
    | Forall (x, t11, t12) ->
        Forall (x, simpl t11 def, t12)
    | Cons (n, t1) ->
        Cons (n, t1)
    | _ -> (app_nf t)
and simpl_listterm l def =
  match l with 
    | [] -> []
    | hd :: tl ->
        (simpl hd def) :: (simpl_listterm tl def)
;;


(*
let rec simpl t def =
(*
  printf "simpl_term: ";
  print_term t;
  printf "\n\n";
*)
  match (app_nf t) with
    | App (t1 , t2) -> (
        match t1 with
          | Lambda (x, t11, t12) -> (
              match t2 with
                | [] -> t1
                | hd :: [] ->
                    let t12' = (rewrite_term_var_term (alpha_term_vars t12 (term_fv hd VarSet.empty)) x hd) in
                      simpl t12' def
                | hd :: tl ->
                    let t12' = (rewrite_term_var_term (alpha_term_vars t12 (term_fv hd VarSet.empty)) x hd) in
                      simpl (App (t12', tl)) def
            )
          | Phi (x, largs, ty, ldecargs, te) -> (
              if (List.length t2 < List.length largs) then
                t
              else (
                let te' = (rewrite_term_var_term te x (Phi (x, largs, ty, ldecargs, te))) in
                let t1' = (largs_lambdas te' largs) in
                let t' = simpl (App (t1', t2)) def in
                  if (nomatch t' VarSet.empty) then
                    t' 
                  else
                    t
              )
            )
          | App (t11, t12) -> (
              simpl (app_nf t) def
            )
          | ( Cste x ) -> (
              match (unfold_term t1 def) with
                | Cons (n, t1') -> App (Cste x, simpl_list t2 def)
                | t1' ->
                    if (t1' = t1) then
                      app_nf (App (t1, simpl_list t2 def))
                    else
                      simpl (App (unfold_term t1' def, simpl_list t2 def)) def
            )
          | Match (t11,ldes, t12) ->
              let t1' = simpl (Match (t11, ldes, t12)) def in
                if (t1 = t1') then
                  (app_nf t)
                else
                  simpl (App (t1', t2)) def                
          | Obj o -> (
              o#exec t2
            )
          | _ -> App (simpl t1 def, simpl_list t2 def)
      )
    | Match (t11,ldes, t12) ->
        let t11' = simpl t11 def in
        let (t11, t12) = (decomp_app (unfold_term t11' def)) in (            
          match t11 with
            | (Cons (n, _)) ->
                simpl (App (List.nth ldes n, t12)) def
            | _ -> (app_nf t)
        )
    | Forall (x, t11, t12) ->
        Forall (x, simpl t11 def, simpl t12 def)
    | Cons (n, t1) ->
        Cons (n, simpl t1 def)
    | _ -> (app_nf t)
and simpl_list l def =
  match l with 
    | [] -> []
    | hd :: tl ->
        (simpl hd def) :: (simpl_list tl def)
;;
*)

let rec unfold_simpl te def =
  (*print_term te; printf "\n"; flush stdout;*)
  let te' = (unfold_term te def) in
    if (te = te') then
      te else      
      let te'' = (simpl te' def) in
        if (te'' = te') then
          te'
        else
          unfold_simpl te'' def
;;

let rec simpl2 t def =
  (*
  printf "simpl2_term: ";
  print_term t VarMap.empty;
  printf "\n\n";
  *)
  match (app_nf t) with
    | App (t1 , t2) -> (
        match t1 with
          | Lambda (x, t11, t12) -> (
              match t2 with
                | [] -> t1
                | hd :: [] ->
                    let t12' = (rewrite_term_var_term (alpha_term_vars t12 (term_fv hd VarSet.empty)) x hd) in
                      simpl2 t12' def
                | hd :: tl ->
                    let t12' = (rewrite_term_var_term (alpha_term_vars t12 (term_fv hd VarSet.empty)) x hd) in
                      simpl2 (App (t12', tl)) def
            )
          | Phi (x, largs, ty, ldecargs, te) -> (
              if (List.length t2 < List.length largs) then
                t
              else (
                let te' = (rewrite_term_var_term te x (Phi (x, largs, ty, ldecargs, te))) in
                let t1' = (largs_lambdas te' largs) in
                let t' = simpl2 (App (t1', t2)) def in
                  if (nomatch t' VarSet.empty) then
                    t' 
                  else
                    t
              )
            )
          | App (t11, t12) -> (
              simpl2 (app_nf t) def
            )
          | ( Cste x ) -> (
              let t1' = (app_nf (unfold_term t1 def)) in
                match t1' with
                  | Lambda (_,_,_) -> 
                      if (List.length t2 >= 1) then
                        simpl2 (App (t1', t2)) def
                      else
                        t
                  | Phi (v, largs, ty, ldecargs, te) -> (
                      if (List.length t2 < List.length largs) then
                        t
                      else (
                        let te' = (rewrite_term_var_term te v (Cste x)) in
                        let t1' = (largs_lambdas te' largs) in
                        let t' = simpl2 (App (t1', t2)) def in
                          if (nomatch t' VarSet.empty) then
                            t' 
                          else (
                            t
                          )
                      )
                    )
                  | Cste c -> simpl2 (App (t1', t2)) def
                  | Obj o -> simpl2 (App (t1', t2)) def
                  | App (t11, t12) -> ( 
                      match t11 with 
                          (*
                        | Lambda (_,_,_) -> 
                            if (List.length (t12 @ t2) >= 1) then
                              simpl2 (App (t11, (t12 @ t2))) def
                            else
                              t
                          *)
                        | Phi (v, largs, ty, ldecargs, te) -> (
                            if (List.length (t12 @ t2) < List.length largs) then
                              t
                            else (
                              let te' = (rewrite_term_var_term te v (Phi (v, largs, ty, ldecargs, te))) in
                              let t1' = (largs_lambdas te' largs) in
                              let t' = simpl2 (App (t1', t12 @ t2)) def in
                                if (nomatch t' VarSet.empty) then
                                  t' 
                                else (
                                  t
                                )
                            )
                          )                          
                        | Cste c -> simpl2 (App (Cste c, t12 @ t2)) def
                        | _ -> App (t1, t2)
                    )
                  | _ -> App (t1, simpl2_listterm t2 def)
            )
          | Match (t11,ldes, t12) ->
              let t1' = simpl2 (Match (t11, ldes, t12)) def in
                if (t1 = t1') then
                  (app_nf t)
                else
                  simpl2 (App (t1', t2)) def                
          | Obj o -> (
              o#exec (simpl2_listterm t2 def)
            )
          | _ -> App (t1, t2)
      )
    | Match (t11,ldes, t12) ->
        let t11' = simpl2 t11 def in
        let (t11, t12') = (decomp_app t11') in (            
          match t11 with
            | (Cons (n, _)) ->
                simpl2 (App (List.nth ldes n, t12')) def
            | Cste c ->
                simpl2 (Match (App (unfold_term (Cste c) def, t12'), ldes, t12)) def
            | _ -> Match (t11', ldes, t12)
        )
    | Forall (x, t11, t12) ->
        Forall (x, simpl2 t11 def, t12)
    | Cons (n, t1) ->
        Cons (n, t1)
    | _ -> (app_nf t)
and simpl2_listterm l def =
  match l with 
    | [] -> []
    | hd :: tl ->
        (simpl2 hd def) :: (simpl2_listterm tl def)
;;
