(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)


open Printf;;
open List;;
open String;;
open Varset;;
open Varmap;;
open Freshness;;
open Def;;
open Term;;
open Rewrite;;
open Alpha;;
open Beta;;
open Pprinter;;
open Termeq;;
open Unification;;
open Listext;;
open Unfold;;
open Fold;;
open Unfoldbetared;;
open Simpl;;

let rec subs_filter s bv =
  match s with
    | [] -> []
    | (hd1, hd2) :: tl ->
        let tl' = subs_filter tl bv in
          if (
            (VarSet.subset (VarSet.singleton hd1) bv) &&
              (VarSet.subset (term_fv hd2 VarSet.empty) bv)
          ) then
            (hd1, hd2) :: tl'
          else
            tl'
;;

let rec subs_filter2 s bv =
  match s with
    | [] -> []
    | (hd1, hd2) :: tl ->
        if (
          (VarSet.subset (VarSet.singleton hd1) bv)) then
          (hd1, hd2) :: subs_filter2 tl (bv +++ (term_fv hd2 bv))
        else
          subs_filter2 tl bv
;;


let rec cutforallprefix_listterm n l =
  match l with
    | [] -> []
    | hd :: tl ->
        let (hdqv, hdc) = decomp_foralls hd in
        let hdqv = nth_tail n hdqv in
          (largs_foralls hdc hdqv) :: (cutforallprefix_listterm n tl)
;;


(* The set of free variables 
   use to assure the freshness

   --> initialise to empty before typechecking a term

*)

type global_fv = {mutable gfv: VarSet.t};;

let tcgfv = {gfv = VarSet.empty}

let rec termcheck 
    (* context *)
    (ctxt: context) 
    (* term to typecheck*)
    (te: term) 
    (* the expected type *)
    (ty: term option) 
    (* the set of bounded variables *)
    (bv: VarSet.t)
    (* The defined constantes *)
    (def: definition) 
    (* the coercions 
       such that (f: A , A)
    *)
    (coercion: (term * term * int) list)     
    (* the oracles *)
    (oracles:  (context -> term -> definition -> term) list)
    (* the overloaded constantes 
       such that 
       x -> (te, ty) :: tl

       tel que x est overloade par te de type ty
       
    *)
    (overload: ((term * int) list) VarMap.t)
    (implicite: (int VarMap.t))
    =  
  (*
    printf "%s |-\n%s: %s\n\n"
    (string_of_listvarterm ctxt VarMap.empty)
    (string_of_term te VarMap.empty)
    (
    match ty with
    | None -> "???"
    | Some ty ->
    (string_of_term ty VarMap.empty)
    ); flush stdout;
  *)
  match (
    match te with


      | Type -> (
          
          match ty with
              (* If we want to guess the type it is simple *)
            | None -> (([], Type, Type, [])::[])
                
            | Some ty -> 
                (* else we try the unification with the desired type*)
                match termcheckerunification ctxt ty Type bv def coercion oracles overload implicite with
                    
                  | Mgu s ->
                      (([], te, apply_subs_term ty (comp_subs s) bv, (comp_subs s))::[])
                        
                  (* Type is always of type Type such that we don't use any coercion *)
                  | _ -> []
        )

      | Obj o -> (

          match ty with
              (* If we want to guess the type it is simple *)
            | None -> (([], Obj o, o#get_type, [])::[])

            | Some ty -> 
                (* else we try the unification with the desired type*)
                match termcheckerunification ctxt (o#get_type) ty bv def coercion oracles overload implicite with
                    
                  | Mgu s ->
                      (([], te, apply_subs_term ty (comp_subs s) bv, (comp_subs s))::[])
                        
                  (* if we cannot unify then we try to coerce *)
                  | _ -> 
                      coerce ctxt (o#get_type) ty bv def coercion oracles overload implicite


        )
      

      | Cste c -> (

          (* find the definition and type of the constante *)
          match find_in_varmap def c with
              
              (* If not found, let's try to find if its an overloading *) 
            | None -> (

                match find_in_varmap overload c with
                    
                  (* No overloading -> this stuff is crap *)
                  | None -> []
                
                  (* This is an overloading, we check each ones as if there was a definition *)
                  | Some l -> (

                      List.fold_left (
                        
                        fun tl' hd ->

                          let (te, i) = hd in
                            
                          let fvs = fresh_names2 (tcgfv.gfv +++ bv +++ (listvarterm2varset ctxt)) i "?T" in
                            tcgfv.gfv <- (list2varset fvs) +++ tcgfv.gfv;
                            
                            (termcheck ctxt (App (te, list_name2list_var fvs)) ty bv def coercion oracles overload implicite) @ tl'
                                    
                      ) [] l


                    )
              )

            (* else we work on the returned type *)
            | Some (_, ctype) -> (

                match ty with
                    
                  (* In the case of inference, it is simple *)
                  | None ->
                      ([], Cste c, ctype, [])::[]

                  (* for the case of typechecking, we use the unification/coercion solution *) 
                  | Some ty ->

                      match termcheckerunification ctxt ctype ty bv def coercion oracles overload implicite with
                          
                        | Mgu s ->
                            ([], Cste c, apply_subs_term ctype (comp_subs s) bv, (comp_subs s))::[]
                              
                        | _ -> 
                            coerce ctxt (Cste c) ty bv def coercion oracles overload implicite
                                  
              )
        )

      | Var v -> (

          (* We first find out if the variable is bound or free *)
          match find_list ctxt v with

            (* It is a free variable *)
            | None -> (

                match ty with
                  | None ->
                      (* For type inference, we just bond it to a type variable (and we bound the type variable to the Type type) *)
                      let fv = 
                        tcgfv.gfv +++ bv                       
                      in
                      let c = fresh_name_list_name fv "?T" in
                        tcgfv.gfv <- c ++ tcgfv.gfv;
                        ((v, Var c)::(c, Type)::[], te, (Var c), [])::[]

                  | Some ty ->
                      (* We assigned it the desired type*)
                      ((v, ty)::[], te, ty, [])::[]          

              )

            (* It is a bound variable *)
            | Some vtype ->

                match ty with

                  (* inference is trivial *)
                  | None ->
                      
                      ([], te, vtype, [])::[]

                  (* for typechecking: unification/coercion *)
                  | Some ty -> (


                      match termcheckerunification ctxt vtype ty bv def coercion oracles overload implicite with

                        | Mgu s ->
                            ([], te, apply_subs_term ty (comp_subs s) bv, (comp_subs s))::[]

                        | _ -> 
                            coerce ctxt te ty bv def coercion oracles overload implicite

                    )                               

        )

      | Lambda (x, t1, t2) -> (

          (* first we verify that t1 is of type Type *)
          let l =  termcheck ctxt t1 (Some Type) bv def coercion oracles overload implicite in
            
            (* The first fold over t1 *)

            List.fold_left (
              
              fun tl' hd -> 
                let (ctxt1, te1, ty1, s1) = hd in

                (* We apply the inference *)
                let t2 = apply_subs_term t2 s1 (x ++ bv) in
                let ctxt =apply_subs_context ctxt s1 bv in
                          

                  (* let's see if we are infering or typechecking *)
                  match ty with

                    (* We are infering *)
                    | None -> (

                        (* we infer t2 knowing that x is bound *)
                        let l = termcheck ((x, te1) :: ctxt1 @ ctxt) t2 None (x ++ bv) def coercion oracles overload implicite in

                          List.fold_left (

                            fun tl' hd ->
                              let (ctxt2, te2, ty2, s2) = hd in
                                
                              (* We apply the inference *)
                              let ctxt1 = apply_subs_context ctxt1 s2 bv in
                              let te1 = apply_subs_term te1 s2 bv in

                                (ctxt2 @ ctxt1, Lambda (x, te1, te2), Forall (x, te1, ty2), s1 @ s2)::tl'



                          ) [] l


                      )

                    (* We are typechecking *)
                    | Some ty -> (

                        (* we apply the inference *)
                        let ty = apply_subs_term ty s1 bv in

                        (* Let's verify that the typecheck is about a Forall *)
                      (*match unfold_beta_red ty def with*)
                          match simpl ty def with
                          (* Yes, we can proceed *)
                          | Forall (y, t3, t4) -> (

                              (* We rewrite t4 such that the Lambda and the Forall quantify the same variable *)
                              let t4 = rewrite_term_var_term t4 y (Var x) in
                                                      
                              (* We apply the inference to t3 and t4 *)
                              let t3 = apply_subs_term t3 s1 bv in
                              let t4 = apply_subs_term t4 s1 bv in
                                
                                (* We typecheck t3 for Type *)
                                
                              let l = termcheck (ctxt1 @ ctxt) t3 (Some Type) bv def coercion oracles overload implicite in

                                List.fold_left (

                                  fun tl' hd ->
                                    
                                    let (ctxt3, te3, ty3, s3) = hd in

                                    (* We apply the inference *)
                                    let ctxt = apply_subs_context ctxt s3 bv in
                                    let ctxt1 = apply_subs_context ctxt1 s3 bv in
                                    let te1 = apply_subs_term te1 s3 bv in
                                    let t2 = apply_subs_term t2 s3 (x ++ bv) in
                                    let t4 = apply_subs_term t4 s3 (x ++ bv) in

                                    (* We typecheck t4 again Type (x being a bounded variable) *)
                                    let l = termcheck ((x, te3) :: ctxt3 @ ctxt1 @ ctxt) t4 (Some Type) (x ++ bv) def coercion oracles overload implicite in
                                      
                                      List.fold_left (
                                        
                                        fun tl' hd ->
                                          
                                          let (ctxt4, te4, ty4, s4) = hd in
                                            
                                          (* We apply the inference *)
                                          let ctxt = apply_subs_context ctxt s4 bv in
                                          let ctxt1 = apply_subs_context ctxt1 s4 bv in
                                          let ctxt3 = apply_subs_context ctxt3 s4 bv in
                                          let te1 = apply_subs_term te1 s4 bv in
                                          let te3 = apply_subs_term te3 s4 bv in
                                          let t2 = apply_subs_term t2 s4 (x ++ bv) in
                                            
                                          (* and finally we typecheck t2 again te4 (x being bounded) *)
                                            
                                          let l = termcheck ((x, te1) :: ctxt4 @ ctxt3 @ ctxt1 @ ctxt) t2 (Some te4)  (x ++ bv) def coercion oracles overload implicite in
                                            
                                            
                                            List.fold_left (
                                              
                                              fun tl' hd ->
                                                
                                                let (ctxt2, te2, ty2, s2) = hd in
                                                  
                                                (* We apply the inference *)
                                                let ctxt = apply_subs_context ctxt s2 bv in
                                                let ctxt1 = apply_subs_context ctxt1 s2 bv in
                                                let ctxt3 = apply_subs_context ctxt3 s2 bv in
                                                let ctxt4 = apply_subs_context ctxt4 s2 bv in
                                                let te1 = apply_subs_term te1 s2 bv in
                                                let te3 = apply_subs_term te3 s2 bv in
                                                let te4 = apply_subs_term te4 s2 (x ++ bv) in

                                                  (* We unify te1 and te3 and apply the unifier to all *) 
                                                  match termcheckerunification (ctxt3 @ ctxt1 @ ctxt) te1 te3 bv def coercion oracles overload implicite with
                                                    | Mgu s13 -> (
                                                        
                                                        let ctxt1 = apply_subs_context ctxt1 s13 bv in
                                                        let ctxt3 = apply_subs_context ctxt3 s13 bv in
                                                        let te1 = apply_subs_term te1 s13 bv in
                                                        let te3 = apply_subs_term te3 s13 bv in
                                                        let te2 = apply_subs_term te2 s13 (x ++ bv) in
                                                        let te4 = apply_subs_term te4 s13 (x ++ bv) in

                          

                                                          (* and we return the result *)
                                                          
                                                          (
                                                            ctxt2 @ ctxt4 @ ctxt3 @ ctxt1,
                                                            Lambda (x, te1, te2),
                                                            Forall (x, te3, te4),
                                                            s1 @ s3 @ s13 @ s4 @ s2
                                                          ) :: tl'

                                                      )
                                                        
                                                    (* If we cannot unify there is a problem *)
                                                    | _ -> 
                                                        tl'

                                                        
                                            ) tl' l
                                              
                                      ) tl' l
                                        
                                ) tl' l

                            )

                          (* Maybe we can unify with the inference (not only for variable ???) *)
                          | ty -> (

                              let l = termcheck (ctxt1 @ ctxt) (Lambda (x, te1, t2)) None bv def coercion oracles overload implicite in

                                List.fold_left (

                                  fun tl' hd ->

                                    let (ctxt', te', ty', s') = hd in

                                    let ctxt1 = apply_subs_context ctxt1 s' bv in
                                    let ty = apply_subs_term ty s' bv in
                                      
                                      match termcheckerunification (ctxt' @ ctxt1 @ ctxt) ty' ty bv def coercion oracles overload implicite with
                                        | Mgu s -> (

                                            let ctxt1 = apply_subs_context ctxt1 s bv in
                                            let ctxt' = apply_subs_context ctxt' s bv in
                                            let te' = apply_subs_term te' s bv in
                                            let ty = apply_subs_term ty s bv in
                                            

                                              (
                                                ctxt' @ ctxt1,
                                                te',
                                                ty,
                                                s1 @ s' @ s
                                              ) :: tl'
                                              
                                            

                                          )

                                        | _ -> tl'
                                      
                                  

                                ) tl' l

                            )
                              

                      )


            ) [] l


        )

      | Forall (x, t1, t2) -> (

          (* First let's verify that t1 has type Type *)

          let l = termcheck ctxt t1 (Some Type) bv def coercion oracles overload implicite in

            List.fold_left (

              fun tl' hd ->

                let (ctxt1, te1, ty1, s1) = hd in
                  
                let t2 = apply_subs_term t2 s1 (x ++ bv) in
                let ctxt = apply_subs_context ctxt s1 bv in                  
                  

                let l = termcheck ((x, te1) :: ctxt1 @ ctxt) t2 (Some Type) (x ++ bv) def coercion oracles overload implicite in
                  List.fold_left (

                    fun tl' hd ->

                      let (ctxt2, te2, ty2, s2) = hd in
                        
                      let ctxt = apply_subs_context ctxt s2 bv in                  
                      let ctxt1 = apply_subs_context ctxt1 s2 bv in                  
                      let te1 = apply_subs_term te1 s2 bv in

                        match ty with
                          | None ->
                              
                              (ctxt2 @ ctxt1, Forall (x, te1, te2), Type, s1 @ s2)::tl'
                                
                          | Some ty ->
                              
                              let ty = apply_subs_term ty (s1 @ s2) bv in
                                
                                match termcheckerunification (ctxt2 @ ctxt1 @ ctxt) ty Type bv def coercion oracles overload implicite with
                                  | Mgu s -> (
                                      
                                      let ctxt2 = apply_subs_context ctxt2 s bv in                  
                                      let ctxt1 = apply_subs_context ctxt1 s bv in                  
                                      let te1 = apply_subs_term te1 s bv in
                                      let te2 = apply_subs_term te2 s (x ++ bv) in
                                        
                                        (ctxt2 @ ctxt1, Forall (x, te1, te2), Type, (s1 @ s2 @ s))::tl'
                                          
                                    )
                                      
                                  | _ -> tl'
                                      
                   
                  ) tl' l


            ) [] l

        )

      | Phi (x, largs, t1, ldec, body) -> (
          
          let l = termchecklargs ctxt largs bv def coercion oracles overload implicite in
            
            List.fold_left (

              fun tl' hd -> 

                let (ctxta, largs, sa) = hd in
                  
                let ctxt = apply_subs_context ctxt sa bv in
                let t1 = apply_subs_term t1 sa ((listvarterm2varset largs) +++ bv) in
                let body = apply_subs_term body sa (x ++ ((listvarterm2varset largs) +++ bv)) in
                  
                let l = termcheck ((rev largs) @ ctxta @ ctxt) t1 (Some Type) ((listvarterm2varset largs) +++ bv) def coercion oracles overload implicite in

                  List.fold_left (

                    fun tl' hd ->

                      let (ctxt1, t1, _, s1) = hd in
                      let ctxt = apply_subs_context ctxt s1 bv in
                      let ctxta = apply_subs_context ctxta s1 bv in
                      let largs = apply_subs_listvarterm largs s1 bv in
                      let body = apply_subs_term body s1 (x ++ ((listvarterm2varset largs) +++ bv)) in

                      let l = termcheck ((x, largs_foralls t1 largs) :: ctxt1 @ (rev largs) @ ctxta @ ctxt) body (Some t1) (x ++ (listvarterm2varset largs) +++ bv) def coercion oracles overload implicite in

                        List.fold_left (

                          fun tl' hd ->

                            let (ctxt2, body, tybody, sbody) = hd in
                            let ctxt = apply_subs_context ctxt sbody bv in
                            let ctxt1 = apply_subs_context ctxt1 sbody bv in
                            let ctxta = apply_subs_context ctxta sbody bv in
                            let largs = apply_subs_listvarterm largs sbody bv in
                            let t1 = apply_subs_term t1 sbody bv in
                                  match ty with
                                    | None ->

                                        (ctxt2 @ ctxt1 @ ctxta, 
                                        Phi (x, largs, t1, ldec, body),
                                        (largs_foralls t1 largs),
                                        comp_subs (sa @ s1 @ sbody))::tl'
                                        
                                    | Some ty ->
                                        let ty = apply_subs_term ty (comp_subs (sa @ s1 @ sbody)) bv in
                                          match 
                                            termcheckerunification 
                                              (ctxt2 @ (x, largs_foralls t1 largs) :: ctxt1 @ (rev largs) @ ctxta @ ctxt) 
                                              (largs_foralls t1 largs)
                                              ty
                                              bv 
                                              def 
                                              coercion
                                              oracles 
                                              overload
                                              implicite
                                          with
                                            | Mgu se -> (
                                                
                                                let ctxta = apply_subs_context ctxta se bv in
                                                let ctxt1 = apply_subs_context ctxt1 se bv in
                                                let ctxt2 = apply_subs_context ctxt2 se bv in
                                                let largs = apply_subs_listvarterm largs se bv in
                                                let t1 = apply_subs_term t1 se bv in
                                                let body = apply_subs_term body se bv in

                                                  (
                                                    ctxt2 @ ctxt1 @ ctxta, 
                                                    Phi (x, largs, t1, ldec, body),
                                                    (largs_foralls t1 largs),
                                                    comp_subs (sa @ s1 @ sbody @ se)
                                                  )::tl'
                                              )
                                                
                                            | _ -> tl'

                        ) tl' l

                  ) tl' l

            ) [] l

        )

      | Ind (x, largs, t1, lcons) -> (

          let l = termchecklargs ctxt largs bv def coercion oracles overload implicite in

            List.fold_left (

              fun tl' hd ->

                let (ctxt1, largs, s) = hd in
                  
                let ctxt = apply_subs_context ctxt s bv in
                let t1 = apply_subs_term t1 s ((listvarterm2varset largs) +++ bv) in
                let lcons = apply_subs_listterm lcons s (x ++ ((listvarterm2varset largs) +++ bv)) in

                let l = termcheck ((rev largs) @ ctxt1 @ ctxt) t1 (Some Type) ((listvarterm2varset largs) +++ bv) def coercion oracles overload implicite in
                  
                  List.fold_left (

                    fun tl' hd ->
                      
                      let (ctxtt1, t1, _, st1) = hd in
                        
                      let ctxt = apply_subs_context ctxt st1 bv in
                      let ctxt1 = apply_subs_context ctxt1 st1 bv in
                      let largs = apply_subs_listvarterm largs st1 bv in
                      let lcons = apply_subs_listterm lcons st1 (x ++ (listvarterm2varset largs) +++ bv) in
                      let lcons = largs_foralls_listterm lcons largs in

                      let l = termchecklcons
                        ((x, largs_foralls t1 largs) :: ctxtt1 @ (rev largs) @ ctxt1 @ ctxt)
                        lcons
                        (list_of_n_elt (Some Type) (List.length lcons))                        
                        (x ++ (listvarterm2varset largs) +++ bv)
                        def
                        coercion oracles overload implicite in

                        List.fold_left (

                          fun tl' hd ->

                            let (ctxtlcons, lcons, _, slcons) = hd in
                              
                            let lcons = cutforallprefix_listterm (List.length largs) lcons in
                            let ctxt = apply_subs_context ctxt slcons bv in
                            let ctxt1 = apply_subs_context ctxt1 slcons bv in
                            let ctxtt1 = apply_subs_context ctxtt1 slcons bv in                                
                            let largs = apply_subs_listvarterm largs slcons bv in
                              
                            (*let lcons = apply_subs_listterm lcons slcons fv in*)
                            let t1 = apply_subs_term t1 slcons bv in
                                match ty with
                                  | None -> (

                                      (
                                        ctxtlcons @ ctxtt1 @ ctxt1,
                                        fold (Ind (x, largs, t1, lcons)) def,
                                        (largs_foralls t1 largs),
                                        comp_subs (s @ st1 @ slcons)
                                      )::tl'

                                    )
                                  | Some ty ->
                                      let ty = apply_subs_term ty (comp_subs (s @ st1 @ slcons)) bv in
                                        
                                      match termcheckerunification 
                                        ((x, largs_foralls t1 largs) :: ctxtlcons @ ctxtt1 @ (rev largs) @ ctxt1 @ ctxt)
                                        (largs_foralls t1 largs)
                                        ty                                          
                                        bv 
                                        def coercion oracles overload implicite
                                        with
                                          | Mgu s' ->
                                              let ctxt1 = apply_subs_context ctxt1 s' bv in
                                              let ctxtt1 = apply_subs_context ctxtt1 s' bv in 
                                              let ctxtlcons = apply_subs_context ctxtlcons s' bv in 
                                              let lcons = apply_subs_listterm lcons s' bv in
                                              let t1 = apply_subs_term t1 s' bv in
                                              let largs = apply_subs_listvarterm largs s' bv in
                                                (
                                                  ctxtlcons @ ctxtt1 @ ctxt1,
                                                  fold (Ind (x, largs, t1, lcons)) def,
                                                  (largs_foralls t1 largs),
                                                  comp_subs (s @ st1 @ slcons @ s')
                                                ) :: tl'
                                          | _ -> tl'
                                              

                        ) tl' l

                  ) tl' l

            ) [] l

        )

      | Cons (n, t1) -> (

          let l = termcheck ctxt t1 None bv def coercion oracles overload implicite in

            List.fold_left (

              fun tl' hd ->

                let (ctxt1, t1, ty1, s1) = hd in 

                  match (unfold_beta_red t1 def) with
                    | Ind (x, largs, ty', lcons) -> (
                        
                        let tye = largs_foralls (rewrite_term_var_term (List.nth lcons n) x t1) largs in
                        let te = Cons (n, t1) in
                        
                          match ty with
                            | None -> 
                                (
                                  ctxt1,
                                  te,
                                  tye,
                                  s1
                                ):: tl'

                            | Some ty -> (
                                
                                let ty = apply_subs_term ty s1 bv in
                                let ctxt = apply_subs_context ctxt s1 bv in
                                  
                                match termcheckerunification (ctxt1 @ ctxt) tye ty bv def coercion oracles overload implicite with                                  
                                  | Mgu s ->
                                      let ctxt1 = apply_subs_context ctxt1 s bv in
                                      let tye = apply_subs_term tye s bv in
                                      let te = apply_subs_term te s bv in
                                        (
                                          ctxt1,
                                          te,
                                          tye,
                                          (s1 @ s)
                                        )::tl'
                                  | _ -> tl'
                              )
                      )
                    | _ -> tl'
                        
            ) [] l
        )

      | Match (x, ldes, dty) -> (

          match dty with
            | None -> (
              
                let l = termcheck ctxt x None bv def coercion oracles overload implicite in
                  
                  List.fold_left (
                    fun tl' hd ->
                      let (ctxtx, x, tyx, sx) = hd in
                      let ldes = apply_subs_listterm ldes sx bv in
                      let ctxt = apply_subs_context ctxt sx bv in
                      let (fct, args) = decomp_app (unfold_beta_red tyx def) in 

                        match fct with
                          | (Ind (ind_name, ind_args, ind_ty, ind_lcons)) -> (
                              
                              let ty' = 
                                match ty with
                                  | None -> Type
                                  | Some ty -> apply_subs_term ty sx bv
                              in                            


                                match destruct_cases (ctxtx @ ctxt) x ty' args 0 ldes ind_lcons ind_name ind_args ind_ty ind_lcons  bv def coercion oracles overload implicite with
                                  | None -> tl'
                                  | Some (l1, l2, l3, l4, sd) ->
                                      
                                      let ctxt = apply_subs_context ctxt sd bv in

                                      match ty with
                                        | None -> (
                                            

                                            let l = termcheckldes ctxt l1 l2 (list_of_n_elt None (List.length l1)) l4 bv def coercion oracles overload implicite in
                                              List.fold_left (
                                                fun tl' hd ->
                                                  let (lctxt, ldes', lty, s) = hd in

                                                  let s = comp_subs (sd @ s) in
                                                    
                                                  (* How to infer the most general type from all the lty ??? *)
                                                  (* Now, the solution is REALLY primitive ... *)
                                                  let ldes = apply_subs_listterm ldes s bv in
                                                  let ctxtx = apply_subs_context ctxtx s bv in
                                                  let ctxt = apply_subs_context ctxt s bv in

                                                    match listtypeeq (lctxt @ ctxtx @ ctxt) (List.hd lty) (List.tl lty) bv def coercion oracles overload implicite with
                                                    | None -> tl'
                                                    | Some ty ->
                                                        (lctxt @ ctxtx, Match (x, ldes, None), ty, (sx @ s))::tl'

                                              ) tl' l
                                                
                                          )

                                        | Some ty -> (
                                            
                                            let ty = apply_subs_term ty (comp_subs (sx @ sd)) bv in
                                              
                                            let l = termcheckldes ctxt l1 l2 (list2listoption l3) l4 bv def coercion oracles overload implicite in
                                                  
                                              List.fold_left (
                                                fun tl' hd ->
                                                  let (lctxt, ldes', lty, s) = hd in

                                                  let s' = comp_subs (sd @ s) in
                                                    
                                                  let ldes = apply_subs_listterm ldes s' bv in
                                                  let ctxtx = apply_subs_context ctxtx s' bv in
                                                  let ty = apply_subs_term ty s bv in
                                                  let x = apply_subs_term x s' bv in
                                                      
                                                    (lctxt @ ctxtx, Match (x, ldes, None), ty, (sx @ sd @ s))::tl'
                                              ) tl' l
                                  )
                            )
                          | _ -> tl'
                  ) [] l
                    
              )

            | Some dty -> (

                let l = termcheck ctxt dty (Some Type) bv def coercion oracles overload implicite in

                List.fold_left (
                  
                  fun tl' hd ->

                    let (ctxtdty, dty, tydty, sdty) = hd in
                      
                    let ctxt = apply_subs_context ctxt sdty bv in
                    let ldes = apply_subs_listterm ldes sdty bv in
                    let x = apply_subs_term x sdty bv in
                        
                      match ty with
                        | None -> (
                            
                            let l = termcheck (ctxtdty @ ctxt) (Match (x, ldes, None)) (Some dty) bv def coercion oracles overload implicite in

                              List.fold_left (
                                fun tl' hd ->
                                  let (ctxt1, te1, ty1, s1) = hd in
                                  let ctxtdty = apply_subs_context ctxtdty s1 bv in
                                  let dty = apply_subs_term dty s1 bv in
                                      
                                    match te1 with
                                      | Match (x, ldes, None) ->
                                          (ctxt1 @ ctxtdty, Match (x, ldes, (Some dty)), dty, (sdty @ s1))::tl'
                                      | _ -> tl'
                                            
                              ) tl' l
                            )
                              
                          | Some ty -> (
                              
                              let ty = apply_subs_term ty sdty bv in
                                
                                match termcheckerunification (ctxtdty @ ctxt) dty ty bv def coercion oracles overload implicite with
                                  | Mgu s -> (
                                      
                                      let ctxtdty = apply_subs_context ctxtdty s bv in
                                      let dty = apply_subs_term dty s bv in
                                      let ldes = apply_subs_listterm ldes s bv in
                                      let x = apply_subs_term x s bv in
                                        
                                      let l = termcheck (ctxtdty @ ctxt) (Match (x, ldes, None)) (Some dty) bv def coercion oracles overload implicite in

                                        List.fold_left (
                                          fun tl' hd ->
                                            let (ctxt1, te1, ty1, s1) = hd in
                                              
                                            let ctxtdty = apply_subs_context ctxtdty s1 bv in
                                            let dty = apply_subs_term dty s1 bv in
                                                
                                              match te1 with
                                                | Match (x, ldes, None) ->
                                                    (ctxt1 @ ctxtdty, Match (x, ldes, (Some dty)), dty, (sdty @ s@ s1))::tl'
                                                | _ -> tl'                                                    
                                        ) tl' l                                                      
                                    )                                      
                                  | _ -> tl'                          

                            )
                ) [] l

              )

        )

      | App (t1, t2) -> (

          let l = termcheck ctxt t1 None bv def coercion oracles overload implicite in

            List.fold_left (

              fun tl' hd ->

                let (ctxt1, t1, ty1, s1) = hd in

                let ctxt = apply_subs_context ctxt s1 bv in
                let t2 = apply_subs_listterm t2 s1 bv in

                let l = termcheckapp (ctxt1 @ ctxt) t1 t2 ty1 bv def coercion oracles overload implicite in

                  List.fold_left (

                    fun tl' hd ->

                      let (ctxt2, t, ty2, s2) = hd in

                        let ctxt1 = apply_subs_context ctxt1 s2 bv in
                        let ctxt = apply_subs_context ctxt s2 bv in

                          match ty with
                            | None ->
                                ((ctxt2 @ ctxt1), t, ty2, s1 @ s2)::tl'

                            | Some ty ->
                                let ty = apply_subs_term ty (comp_subs (s1 @ s2)) bv in
                                  
                                  match termcheckerunification (ctxt2 @ ctxt1 @ ctxt) ty2 ty bv def coercion oracles overload implicite with
                                  | Mgu s -> (
                                      
                                      let ctxt1 = apply_subs_context ctxt1 s bv in
                                      let ctxt2 = apply_subs_context ctxt2 s bv in
                                      let t = apply_subs_term t s bv in
                                      let ty = apply_subs_term ty s bv in
                                        
                                        ((ctxt2 @ ctxt1), t, ty, s1 @ s2 @ s)::tl'

                                    )                                      
                                  | _ -> 
                                      let l = coerce ctxt t ty bv def coercion oracles overload implicite in
                                        
                                        List.fold_left (

                                          fun tl' hd ->

                                            let (ctxt3, t3, ty3, s3) = hd in
                                            let ctxt1 = apply_subs_context ctxt1 s3 bv in
                                            let ctxt2 = apply_subs_context ctxt2 s3 bv in

                                              ((ctxt3 @ ctxt2 @ ctxt1), t3, ty3, s1 @ s2 @ s3)::tl'
                                          
                                        )  tl' l

                  ) tl' l                  

            ) [] l

        )

  ) with
    | l -> 
        if (List.length l = 0) then (
          (*
          printf "--------\nError\n--------\n%s|-\n%s: %s\n-------\n\n"
            (string_of_listvarterm ctxt VarMap.empty)
            (string_of_term te VarMap.empty)
            (match ty with
              | None -> "??"
              | Some ty -> string_of_term ty VarMap.empty
            );
          *)
          []
        ) else
          List.map (
            fun hd ->
              let (ctxt, te, ty, s) = hd in
                (*
                printf "********\n%s|-\n%s: %s\nsubst = %s\n*********\n\n"
                  (string_of_listvarterm ctxt VarMap.empty)
                  (string_of_term te VarMap.empty)
                  (string_of_term ty VarMap.empty)
                  (string_of_listvarterm (comp_subs s) VarMap.empty);
                *)
                (*(subs_filter2 ctxt ((term_fv te VarSet.empty) +++ (term_fv ty VarSet.empty)), te, ty, (comp_subs s))*)
                (ctxt, te, simpl ty def, (comp_subs s))
                  
          ) l

(********************************************)
(* Typechecking for mutually recursive term *)
(********************************************)

and termcheckmutual ctxt n lname lte lty bv def coercion oracles overload implicite =
  match (lte, lty) with
    | ([],[]) -> ([],[],[],[])::[]
    | (hdte::tlte, hdty::tlty) -> (

        let bv' = (list2varset ((nth_head n lname) @ (nth_tail (n+1) lname))) in

        let l = termcheck ctxt hdte hdty (bv +++ bv') def coercion oracles overload implicite in

          List.fold_left (

            fun tl' hd ->

              let (hdctxt, hdte, hdty, s') = hd in

              let ctxt = apply_subs_context ctxt s' bv in
              let tlte = apply_subs_listterm tlte s' bv in
              let tlty = apply_subs_listoptionterm tlty s' bv in
                
              let s2 =

                match find_list ctxt (List.nth lname n) with
                  | None -> []
                  | Some hdty' ->

                      match termcheckerunification (hdctxt @ ctxt) hdty' hdty bv def coercion oracles overload implicite with
                        | Mgu s -> s
                        | _ -> []

              in

              let ctxt = apply_subs_context ctxt s2 bv in
              let hdctxt = apply_subs_context hdctxt s2 bv in
              let tlte = apply_subs_listterm tlte s2 bv in
              let tlty = apply_subs_listoptionterm tlty s2 bv in
              let hdte = apply_subs_term hdte s2 bv in
              let hdty = apply_subs_term hdty s2 bv in



                
              let l = termcheckmutual (hdctxt @ ctxt) (n+1) lname tlte tlty bv def coercion oracles overload implicite in

                List.fold_left (

                  fun tl' hd ->

                    let (tlctxt, tlte, tlty, s'') = hd in

                    let hdctxt = apply_subs_context hdctxt s'' bv in
                    let hdte = apply_subs_term hdte s'' bv in
                    let hdty = apply_subs_term hdty s'' bv in

                      (tlctxt @ hdctxt, hdte::tlte, hdty::tlty, comp_subs (s' @ s2 @ s'')) :: tl'


                ) tl' l



          ) [] l


      )
    | _ -> []






and termcheckapp ctxt te args ty bv def coercion oracles overload implicite =
  match args with
    | [] -> ([], app_nf te, ty, [])::[]
    | hd :: tl -> (

        (*let ty = unfold_beta_red ty def in*)
        let ty = simpl ty def in
          
          match ty with
            | Forall (x, ty1, ty2) -> (

                let l = termcheck ctxt hd (Some ty1) bv def coercion oracles overload implicite in
                  
                  List.fold_left (

                    fun tl' hd ->
                      
                      let (hdctxt, hdte, hdty, hds) = hd in
                        
                      let ctxt = apply_subs_context ctxt hds bv in
                      let te = apply_subs_term te hds bv in
                      let tl = apply_subs_listterm tl hds bv in
                      let ty2 = apply_subs_term ty2 hds (x ++ bv) in

                      let ty = (rewrite_term_var_term (alpha_term_vars ty2 (term_fv hdte VarSet.empty)) x hdte) in
                        
                      let l = termcheckapp (hdctxt @ ctxt) (App (te, hdte::[])) tl ty bv def coercion oracles overload implicite in

                        List.fold_left (

                          fun tl' hd ->
                            
                            let (tlctxt, tlte, tlty, tls) = hd in
                            let hdctxt = apply_subs_context hdctxt tls bv in
                              
                              (tlctxt @ hdctxt, tlte, tlty, comp_subs (hds @ tls))::tl'

                        ) tl' l                         

                  ) [] l

              )

            | Var x -> (

                if not (vsin x bv) then (

                  let v = fresh_name_list_name (tcgfv.gfv +++ bv) "?V" in
                    tcgfv.gfv <- v ++ tcgfv.gfv;
                    let x1 = fresh_name_list_name (tcgfv.gfv +++ bv) "?T" in
                      tcgfv.gfv <- x1 ++ tcgfv.gfv;
                      let ty1 = Var x1 in
                      let x2 = fresh_name_list_name (tcgfv.gfv +++ bv) "?T" in
                        tcgfv.gfv <- x2 ++ tcgfv.gfv;
                        let ty2 = Var x2 in
                        let sx = (x, Forall (v, ty1, ty2))::[] in

                        let ctxt' = (x2, Type)::(x1, Type)::[] in
                        let ctxt = apply_subs_context ctxt sx bv in
                        let te = apply_subs_term te sx bv in
                        let hd = apply_subs_term hd sx bv in
                        let tl = apply_subs_listterm tl sx bv in
                          
                        let l = termcheck (ctxt' @ ctxt) hd (Some ty1) bv def coercion oracles overload implicite in
                          
                          List.fold_left (
                            
                            fun tl' hd ->
                              
                              let (hdctxt, hdte, hdty, hds) = hd in

                                
                              let ctxt = apply_subs_context ctxt hds bv in
                              let ctxt' = apply_subs_context ctxt' hds bv in
                              let te = apply_subs_term te hds bv in
                              let tl = apply_subs_listterm tl hds bv in
                              let ty2 = apply_subs_term ty2 hds (v ++ bv) in
                                
                              let ty = (rewrite_term_var_term (alpha_term_vars ty2 (term_fv hdte VarSet.empty)) v hdte) in
                                
                              let l = termcheckapp (hdctxt @ ctxt' @ ctxt) (App (te, hdte::[])) tl ty bv def coercion oracles overload implicite in
                                
                                List.fold_left (
                                  
                                  fun tl' hd ->
                                    
                                    let (tlctxt, tlte, tlty, tls) = hd in
                                    let hdctxt = apply_subs_context hdctxt tls bv in
                                    let ctxt' = apply_subs_context ctxt' tls bv in
                                      
                                      (tlctxt @ hdctxt @ ctxt', tlte, tlty, comp_subs (sx @ hds @ tls))::tl'
                                        
                                ) tl' l                         
                                  
                          ) [] l

                ) else (

                  (*
                  printf "prout1!!!!\n%s \\in %s\n" x (string_of_listofvar (VarSet.elements bv));
                  *)
                  
                  
                  []

                )
              )

            | _ -> 
                []

      )

and termcheckldes ctxt lctxt lcons ltype lbv bv def coercion oracles overload implicite =
  match (lctxt, lcons, ltype, lbv) with
    | ([],[],[],[]) -> ([],[],[],[])::[]
    | (hdctxt::tlctxt, hdte::tlte, hdty::tlty, hdbv::tlbv) ->(
	(*
        printf "termcheckldes:\n%s|-\n%s: %s\n\n"
          (string_of_listvarterm hdctxt VarMap.empty)
          (string_of_term hdte VarMap.empty)
	  (match hdty with
	     | None -> "??"
	     | Some hdty ->
		 (string_of_term hdty VarMap.empty)
	  );
	*)

        let l = termcheck (hdctxt) hdte hdty (hdbv +++ bv) def coercion oracles overload implicite in
          
          List.fold_left (
            
            fun tl' hd ->
              
              let (ctxt', hd', ty', s') = hd in
                
              let ctxt = apply_subs_context ctxt s' bv in
              let tlctxt = apply_subs_listcontext tlctxt s' bv in
              let tlte = apply_subs_listterm tlte s' bv in
              let tlty = apply_subs_listoptionterm tlty s' bv in
                
              let l = termcheckldes ctxt tlctxt tlte tlty tlbv bv def coercion oracles overload implicite in
                
                List.fold_left (
                  
                  fun tl' hd ->
                    
                    let (ctxt'', tlc'', ty'', s'') = hd in
                    let ctxt' = apply_subs_context ctxt' s'' bv in
                    let hd' = apply_subs_term hd' s'' bv in
                    let ty' = apply_subs_term ty' s'' bv in
                      
                      ( ctxt'' @ ctxt', hd'::tlc'', ty'::ty'', comp_subs (s'@s''))::tl'
                      
                        
                        
                ) tl' l                
                  
                  
          ) [] l

      )
    | _ -> (* probleme *) []


and termchecklcons ctxt lcons ltype bv def coercion oracles overload implicite =
  match (lcons, ltype) with
    | ([], []) -> ([],[],[],[])::[]
    | (hdc :: tlc, hdty::tlty) -> (

        let l = termcheck ctxt hdc hdty bv def coercion oracles overload implicite in

          List.fold_left (

            fun tl' hd ->

              let (ctxt', hd', ty', s') = hd in
                
              let ctxt = apply_subs_context ctxt s' bv in
              let tlc = apply_subs_listterm tlc s' bv in
              let tlty = apply_subs_listoptionterm tlty s' bv in

              let l = termchecklcons ctxt tlc tlty bv def coercion oracles overload implicite in

                List.fold_left (

                  fun tl' hd ->

                    let (ctxt'', tlc'', ty'', s'') = hd in
                    let ctxt' = apply_subs_context ctxt' s'' bv in
                    let hd' = apply_subs_term hd' s'' bv in
                    let ty' = apply_subs_term ty' s'' bv in
                      
                      (ctxt'' @ ctxt', hd'::tlc'', ty'::ty'', comp_subs (s' @ s''))::tl'                      

                ) tl' l                
                

          ) [] l
      )
    | _ -> (* probleme *) []

and termchecklargs ctxt largs bv def coercion oracles overload implicite =
  match largs with
    | [] -> ([],[],[])::[]
    | (hd1, hd2) :: tl ->
        let l = termcheck ctxt hd2 (Some Type) bv def coercion oracles overload implicite in
          
          List.fold_left (

            fun tl' hd ->
              let (ctxt', hd2, _, s') = hd in
                
              let ctxt = apply_subs_context ctxt s' bv in
              let tl = apply_subs_listvarterm tl s' bv in
                
              let l = termchecklargs ((hd1, hd2) :: ctxt' @ ctxt) tl (hd1 ++ bv) def coercion oracles overload implicite in
                
                List.fold_left (
                  
                  fun tl' hd ->
                    
                    let (ctxt'', tl'', s'') = hd in
                      
                    let ctxt' = apply_subs_context ctxt' s'' bv in
                    let hd2 = apply_subs_term hd2 s'' bv in

                      (ctxt'' @ ctxt', (hd1, hd2):: tl'', comp_subs (s' @ s''))::tl'
                      
                      
                ) tl' l
                  
                  
          ) [] l

(**********************************)
(* term unification using oracles *)
(**********************************)

and termcheckerunification ctxt te1 te2 bv def coercion oracles overload implicite =
  let te1 = (simpl te1 def) in
  let te2 = (simpl te2 def) in
  (* te1 and te2 must be well-typed *)
  match unification_term te1 te2 bv def with
    | NoMgu ->
        NoMgu
    | DnkMgu -> 
        (* generalize with a unification that use oracles*)
        
        (* special case: equality *)
        (
          match termcheck ctxt te1 None bv def coercion oracles overload implicite with
          | [] -> NoMgu
          | (ctxt1, te1, ty1, s1)::tl ->
              
              let eq = Ind ("eq", ("A", Type) :: ("a", Var "A") :: [], Forall ("b", Var "A", Type), (App (Var "eq", (Var "A")::(Var "a")::(Var "a")::[])) :: []) in
              let prop = App (eq, ty1 :: te1 :: te2 ::[]) in
                if 
                  termcheckeqoracles ctxt prop bv def coercion oracles overload oracles implicite
                then 
                  Mgu []
                else
                  DnkMgu
              
        )

    | Mgu s ->
        Mgu s

and termcheckerunificationlist ctxt l1 l2 bv def coercion oracles overload implicite =
  match (l1, l2) with
    | ([], []) -> Mgu []
    | (hd1 :: tl1, hd2 :: tl2) -> (
        
        match termcheckerunification ctxt hd1 hd2 bv def coercion oracles overload implicite with
          | NoMgu -> NoMgu
          | DnkMgu -> DnkMgu
          | Mgu s ->
              let ctxt = apply_subs_context ctxt s bv in
              let tl1 = apply_subs_listterm tl1 s bv in 
              let tl2 = apply_subs_listterm tl2 s bv in 
                match termcheckerunificationlist ctxt tl1 tl2 bv def coercion oracles overload implicite with
                  | NoMgu -> NoMgu
                  | DnkMgu -> DnkMgu
                  | Mgu s' ->
                      Mgu (comp_subs (s @ s'))
                      

      )
    | _ -> DnkMgu


(******************)
(* Really bad !!! *)
(******************)

and listtypeeq ctxt ty lt bv def coercion oracles overload implicite =

  List.fold_left (

    fun tl' hd ->
      
      if (term_eq (simpl ty def) (simpl hd def) def) then
        tl'
      else
        None              

  ) (Some ty) lt
  

(**********************************************************************)
(* Equality of term, through classical equality, or through an oracle *)
(* te1 and te2 must be well-typed                                     *)
(**********************************************************************)

and termcheckeq ctxt te1 te2 ty bv def coercion oracles overload implicite =
  if (term_eq (simpl te1 def) (simpl te2 def) def) then
    (* the answer got through the *)
    true
  else
    let eq = Ind ("eq", ("A", Type) :: ("a", Var "A") :: [], Forall ("b", Var "A", Type), (App (Var "eq", (Var "A")::(Var "a")::(Var "a")::[])) :: []) in
    let prop = App (eq, ty :: te1 :: te2 ::[]) in
      termcheckeqoracles ctxt prop bv def coercion oracles overload oracles implicite  

(********************)
(* calls to oracles *)
(********************)      

and termcheckeqoracles ctxt prop bv def coercion oracles overload lo implicite =
  match lo with
    | [] -> false
    | hd :: tl ->
        let proof = hd ctxt prop def in
          match termcheck ctxt prop (Some proof) bv def coercion oracles overload implicite with
            | [] -> termcheckeqoracles ctxt prop bv def coercion oracles overload tl implicite
            | _ -> true

(******************)
(*  coercion      *)
(******************)

(* First, we coerce only on simple types (without variables) *)

and coerce ctxt te ty bv def coercion oracles overload implicite =
  
  List.fold_left (
    
    fun acc hd ->
      
      let (te', ty', i) = hd in

      let ty' = alpha_term_vars ty' bv in

      let (_, ccl) = decomp_foralls ty' in

        match termcheckerunification ctxt ccl ty bv def coercion oracles overload implicite with
          | NoMgu -> acc
          | DnkMgu -> acc
          | Mgu _ ->                
        
              let fvs = fresh_names2 (tcgfv.gfv +++ bv +++ (listvarterm2varset ctxt)) i "?T" in
                tcgfv.gfv <- (list2varset fvs) +++ tcgfv.gfv;
                
                let nte = App (te', (list_name2list_var fvs) @ te::[]) in
                  
                let nte = alpha_term_vars nte bv in
                  
                  acc @ (termcheck ctxt nte (Some ty) bv def coercion oracles overload implicite)

  ) [] coercion 
    

    
  

(************************)
(* Destruction of cases *)
(************************)

and destruct_cases ctxt te ty args n ldes lcons ind_name ind_args ind_ty ind_lcons bv def coercion oracles overload implicite =
  match lcons with
    | [] -> (
        match ldes with
          | [] -> Some ([], [], [], [], [])
          | _ -> printf "destruct_case1\n"; None
      )
    | hdc :: tlc ->
        let ind = fold (Ind (ind_name, ind_args, ind_ty, ind_lcons)) def in
        let hdc = (alpha_term_vars (largs_foralls (rewrite_term_var_term hdc ind_name ind) ind_args) ((listvarterm2varset ctxt) +++ (listterm_fv args VarSet.empty) +++ bv)) in
        let (hdcqv, hdccon) = (decomp_foralls hdc) in 
        let (hdcconfct, hdcconargs) = (decomp_app hdccon) in 
          
          match termcheckerunificationlist ctxt hdcconargs args
            (*fv*)
            (VarSet.diff bv (listterm_fv args VarSet.empty))
            def coercion oracles overload implicite
          with
            | NoMgu ->
                destruct_cases ctxt te ty args (n+1) ldes tlc ind_name ind_args ind_ty ind_lcons bv def coercion oracles overload implicite
            | a ->
                match ldes with
                  | [] -> printf "destruct_case2 (%s Vs %s := %s)\n" (string_of_listterm args implicite) (string_of_listterm hdcconargs implicite) (
                      match a with
                        | Mgu s -> (string_of_listvarterm s implicite)
                        | NoMgu -> "NoMgu :D"
                        | DnkMgu -> "DnkMgu :D"
                    ); None
                  | hdd :: tld ->

                      match destruct_cases ctxt te ty args (n+1) tld tlc ind_name ind_args ind_ty ind_lcons bv def coercion oracles overload implicite with
                        | None -> None
                        | Some (l1, l2, l3, l4, s'') ->
                            
			    (*
                            printf "(%s Vs %s := %s)\n" (string_of_listterm hdcconargs implicite) (string_of_listterm args implicite) (
                              match a with
                              | Mgu s -> (string_of_listvarterm s implicite)
                              | NoMgu -> "NoMgu :D"
                              | DnkMgu -> "DnkMgu :D"
                              ); 
                            *)
                            let s = (
                              match a with
                                | Mgu s -> (comp_subs s)
                                | _ -> []
                            ) in
                              
                              let ctxt = (rev hdcqv) @ ctxt in                        
                              let ctxt = apply_subs_context ctxt s VarSet.empty in
                              let hdd = (alpha_term_vars hdd (listvarterm2varset ctxt)) in    
                              let p = (app_nf (App (Cons (n,ind), apply_subs_listterm (list_name2list_var (list_proj1 hdcqv)) s VarSet.empty))) in
                              let ctxt = rewrite_context_term_term ctxt te p (*fv*) VarSet.empty def in 
                              let ty = rewrite_term_term_term ty te p (*fv*) VarSet.empty def in
                              let ty = apply_subs_term ty s VarSet.empty in
                              let hdd = rewrite_term_term_term hdd te p VarSet.empty def in
                              let hdd = apply_subs_term hdd s VarSet.empty in
                              let (hddlam,_) = decomp_nlambdas hdd (List.length hdcqv) in
                              let hdd = app_nf (App (hdd, apply_subs_listterm (list_name2list_var (list_proj1 hdcqv)) s VarSet.empty)) in
                                
                              let lt1 = (
                                rev (list_proj2 (rewrite_context_term_term (apply_subs_context (rev hdcqv) s VarSet.empty) te p VarSet.empty def))
                              ) in
                              let lt2 = (list_proj2 hddlam) in
                                (*
                                printf "%s Vs %s\n"
                                  (string_of_listterm lt1 VarMap.empty)
                                  (string_of_listterm lt2 VarMap.empty);
                                *)

                                let s' = (

                                  
                                  match termcheckerunificationlist ctxt lt2 lt1
                                    (*fv*)
                                    (VarSet.diff bv (listterm_fv args VarSet.empty))
                                    def coercion oracles overload implicite with
                                      | Mgu s -> s
                                      | _ -> []
                                  
                                ) in
                                let hdd = simpl hdd def in
                                  (*
                                  printf "%s|-\n%s: %s\n"
                                    (string_of_listvarterm ctxt VarMap.empty)
                                    (string_of_term hdd VarMap.empty)
                                    (string_of_term ty VarMap.empty);
                                  printf "[%s]\n\n" (string_of_listvarterm s VarMap.empty);
                                  *)
                                Some (ctxt::l1, hdd::l2, ty::l3, (list2varset (list_proj1 hdcqv))::l4, comp_subs (s'' @ s'))


;;
  
