(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open Def;;
open Pprinter;;
open State;;
open String;;
open Simpl;;
open Varmap;;

(* The class for the bool type *)

class boolTypeObj =
object (self)
  inherit [term] pObj
  method get_name = "Bool"
  method get_type = Type
  method pprint = (fun args -> self#get_name)
  method my_self = (self :> term pObj)
  method exec = (fun args -> (Obj (self#my_self)))
  method show = self#get_name
end;;

let boolType = 
  let o = new boolTypeObj in
    Obj o#my_self
;;

(* The class for the bool term *)
class boolTermObj b =
object (self)
  inherit [term] pObj
  val mutable boolvalue = b
  method get_boolvalue = boolvalue
  method get_name = string_of_bool boolvalue
  method get_type = boolType
  method pprint = (fun args -> self#get_name)
  method my_self = (self :> term pObj)
  method exec = (fun args -> (Obj (self#my_self)))
  method show = self#get_name
end;;

let create_boolTerm b =
  let o = new boolTermObj b in
    Obj o#my_self
;;

(* The class for the negation function *)
class bnegTermObj  =
object (self)
  inherit [term] pObj
  method get_name = "bneg"
  method get_type = Forall ("H0", boolType, boolType)
  method pprint = (fun args -> 
    match args with
      | [] -> "!_"
      | hd::tl ->
          concat "" ("! " :: (string_of_term hd state.implicite) :: [])
  )
  method my_self = (self :> term pObj)
  method exec = (fun args -> 
    
    match (simpl_listterm args state.def) with
      | (Obj o1) :: [] -> (

          let o1 = ((Obj.magic o1) :> boolTermObj) in
            create_boolTerm (not o1#get_boolvalue)
        )       
      | a -> App (Obj self#my_self, a)


  )
  method show = self#get_name
end;;

let bnegTerm =
  let o = new bnegTermObj in
    Obj o#my_self
;;



(* The class for the conjunction function *)
class bandTermObj  =
object (self)
  inherit [term] pObj
  method get_name = "band"
  method get_type = Forall ("H0", boolType, Forall ("H0", boolType, boolType))
  method pprint = (fun args -> 
    match args with
      | [] -> "_&&_"
      | hd::tl ->
          concat "" ((string_of_term hd state.implicite) :: " &&" ::
                        (
                          match tl with
                            | [] -> "_"
                            | hd :: tl ->
                                concat "" (" " :: (string_of_term hd state.implicite) :: [])
                        ) :: []
          )
  )
  method my_self = (self :> term pObj)
  method exec = (fun args -> 
    
    match (simpl_listterm args state.def) with
      | (Obj o1) :: (Obj o2) :: [] -> (

          let o1 = ((Obj.magic o1) :> boolTermObj) in
          let o2 = ((Obj.magic o2) :> boolTermObj) in
            create_boolTerm (o1#get_boolvalue && o2#get_boolvalue)
        )       
      | a -> App (Obj self#my_self, a)


  )
  method show = self#get_name
end;;

let bandTerm =
  let o = new bandTermObj in
    Obj o#my_self
;;

(* The class for the disjunction function *)
class borTermObj  =
object (self)
  inherit [term] pObj
  method get_name = "bor"
  method get_type = Forall ("H0", boolType, Forall ("H0", boolType, boolType))
  method pprint = (fun args -> 
    match args with
      | [] -> "_||_"
      | hd::tl ->
          concat "" ((string_of_term hd state.implicite) :: " ||" ::
                        (
                          match tl with
                            | [] -> "_"
                            | hd :: tl ->
                                concat "" (" " :: (string_of_term hd state.implicite) :: [])
                        ) :: []
          )
  )
  method my_self = (self :> term pObj)
  method exec = (fun args -> 
    
    match (simpl_listterm args state.def) with
      | (Obj o1) :: (Obj o2) :: [] -> (

          let o1 = Obj.magic o1 in
          let o2 = Obj.magic o2 in
            create_boolTerm (o1#get_boolvalue || o2#get_boolvalue)
        )       
      | a -> App (Obj self#my_self, a)


  )
  method show = self#get_name
end;;

let borTerm =
  let o = new borTermObj in
    Obj o#my_self
;;

(* The class for the equality function *)
class beqTermObj  =
object (self)
  inherit [term] pObj
  method get_name = "beq"
  method get_type = Forall ("H0", boolType, Forall ("H0", boolType, boolType))
  method pprint = (fun args -> 
    match args with
      | [] -> "_==_"
      | hd::tl ->
          concat "" ((string_of_term hd state.implicite) :: " ==" ::
                        (
                          match tl with
                            | [] -> "_"
                            | hd :: tl ->
                                concat "" (" " :: (string_of_term hd state.implicite) :: [])
                        ) :: []
          )
  )
  method my_self = (self :> term pObj)
  method exec = (fun args -> 
    
    match (simpl_listterm args state.def) with
      | (Obj o1) :: (Obj o2) :: [] -> (

          let o1 = Obj.magic o1 in
          let o2 = Obj.magic o2 in
            create_boolTerm (o1#get_boolvalue = o2#get_boolvalue)
        )       
      | a -> App (Obj self#my_self, a)


  )
  method show = self#get_name
end;;

let beqTerm =
  let o = new beqTermObj in
    Obj o#my_self
;;

(* The class for the ifte function *)
class bifteTermObj  =
object (self)
  inherit [term] pObj
  method get_name = "bifte"
  method get_type = Forall ("H0", boolType, Forall ("H0", boolType, boolType))
  method pprint = (fun args -> 
    match args with
      | [] -> "_==_"
      | hd::tl ->
          concat "" ((string_of_term hd state.implicite) :: " ==" ::
                        (
                          match tl with
                            | [] -> "_"
                            | hd :: tl ->
                                concat "" (" " :: (string_of_term hd state.implicite) :: [])
                        ) :: []
          )
  )
  method my_self = (self :> term pObj)
  method exec = (fun args -> 
    
    match (simpl_listterm args state.def) with
      | (Obj o1) :: (Obj o2) :: [] -> (

          let o1 = Obj.magic o1 in
          let o2 = Obj.magic o2 in
            create_boolTerm (o1#get_boolvalue = o2#get_boolvalue)
        )       
      | a -> App (Obj self#my_self, a)


  )
  method show = self#get_name
end;;

let beqTerm =
  let o = new beqTermObj in
    Obj o#my_self
;;

(* Initialisation *)
let bool_init =  
  add_overload "eq" beqTerm 0
;;
