%{

(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

  open Printf;;
  open Lexing;;
  open List;;
  open String;;
  
  open Def;;
  open Pprinter;;
  open Freshness;;
  open Term;;
  open Alpha;;
  open Beta;;
  open Termeq;;
  open Varset;;
  open Varmap;;
  open Arith;;
  open Terminaison;;
  open Rewrite;;
  open Fold;;
  open Unfold;;
  open Simpl;;
  open Listext;;
  open Termchecker;;
  open Kernel;;
  open Listext;;

  open Indprinc;;
  open Tactic;;

  (* *)
  
  type pcmd =
      | Definition of (string) list
      | Command
      | Coercion
      | Overload of (string * term)
      | Implicite of string
      | Lemma of string
      | Tactic of tactics
  ;;


  type proof_tree =
    (((context * string * term) list) * (string * term * term)) list;;  

  type int_state = {
    mutable def: definition; 
    mutable coercion: (term * term * int) list;
    mutable oracles:  (context -> term -> definition -> term) list;
    mutable overload: ((term * int) list) VarMap.t;
    mutable pt: proof_tree; 
    mutable lastcmd: pcmd list; 
    mutable implicite: int VarMap.t;
    mutable subgoals: int;
  };;

  let state : int_state = {

    def = VarMap.empty; 
    coercion = [];
    oracles = [];
    overload = VarMap.empty;
    pt = []; 
    lastcmd = []; 
    implicite = VarMap.empty;
    subgoals = 0;
  };;

  let proof_mode() =
    match state.pt with
      | [] -> false
      | _ -> true
  ;;


  let rec print_subgoal_context c =
    match c with
      | [] -> printf "";
      | (hd1, hd2) :: tl ->
          printf "%s : %s\n" hd1 (string_of_term hd2 state.implicite);
          print_subgoal_context tl;;
  

  let print_subgoal ctxt ty =
    printf "\n\n";

    printf "\n\n%s----------------------------\n\n%s\n\n"
      (string_of_listvarterm ctxt state.implicite)
      (string_of_term ty state.implicite);          
  ;;

  let rec print_subgoals sg =
    match sg with
      | [] -> printf "\n\n";
      | (hd1, hd2, hd3) :: tl ->
          printf "----------------------------\n%s\n\n"          
            (string_of_term hd3 state.implicite);
          print_subgoals tl
  ;;

  let print_subgoals_ctxt () =
    match (state.pt) with
      | [] -> printf "Error: Not in proof mode\n\n";
      | ((hd1,hd2,hd3)::tl,(ln,te,ty)) :: tl2 ->  (

          printf "\n\n";
          (*print_subgoal_context (readable_context_fp hd1 state.def);*)
          print_subgoal_context hd1;
          print_subgoals ((hd1,hd2,hd3)::tl);
          flush stdout;
        
        );
      | ([], g) :: tl -> printf "No more subgoals\n\n";;


  let rec remove_all_tactic l =
    match l with
      | [] -> []
      | hd :: tl ->
          match hd with
            | Tactic s ->
                remove_all_tactic tl
            | _ ->
                l
  ;;
  
  class axiom my_name my_type =
  object (self)
    inherit [term] pObj
    method get_name = my_name
    method get_type = my_type
    method pprint = (fun args -> concat "" (self#get_name :: (string_of_listterm args state.implicite) :: []))
    method my_self = (self :> term pObj)
    method exec = (fun args -> App (Obj (self#my_self), simpl_listterm args state.def))
    method show = (string_of_term self#get_type state.implicite)
  end;;

%}

/*  */
%token NEWLINE EOF

/* The indentifiers variables */
%token <string> NAME

/* Integer (for indexes of contructors) */
%token <int> NUM

%token UNDERSCORE COERCE AROBAS

/* quantifier */
%token LAMBDA IND PHI FORALL
  
/* term */
%token MATCH RETURN WITH CONS OF TYPE IMPL

%right IMPL

/* parenthese */
%token LPAREN RPAREN LBRACK RBRACK LACCOL RACCOL

/* ponctuation */
%token PAR SCOMMA COMMA DOT DDOT TURNSTIL EQ 
  
%left SCOMMA
%right PAR

/* command */ 
%token ALPHA TYPECHECK QUIT SIMPL SHOWDEF BETA SHOWCOERCION UNDO LEMMA DROP QED DEFINED

/* definition */ 
%token INDUCTIVE RECURSIVE DEFINITION COERCION OVERLOAD IMPLICITE

/* tactics */
%token EQUIV INTRO INTROS GENERALIZE AS APPLY APPLY2 EXACT SIMPL IN FACTORIZE DESTRUCT DEFINED UNFOLD FAIL IDTAC REPEAT DO TRY

%start input command
%type <unit> input command
  
%% /* Grammar rules and actions follow */

  input: 
    
| command input {}

| definition input {}
        
| EOF {}
          
    
  ;               
  
  command :

| ALPHA term DOT {
    
    let t = $2 in
    let t = alpha_term_vars t (vmdomain state.def) in
    let t = term_diff t VarSet.empty VarSet.empty in
      state.lastcmd <- Command :: state.lastcmd;
      print_term t VarMap.empty; printf "\n\n";

  }

| TYPECHECK term DOT {

    match check_term $2 None state.def state.coercion state.oracles state.overload state.implicite false with
      | None -> printf "Cannot Type Check\n\n";
      | Some (te, ty) ->
          state.lastcmd <- Command :: state.lastcmd;
          printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);


  }

| TYPECHECK term DDOT term DOT {

    match check_term $2 (Some $4) state.def state.coercion state.oracles state.overload state.implicite false with
      | None -> printf "Cannot Type Check\n\n";
      | Some (te, ty) ->
          state.lastcmd <- Command :: state.lastcmd;
          printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);

  }

| SIMPL term DOT {

    match simpl_term $2 state.def state.coercion state.oracles state.overload state.implicite with
      | None -> printf "Cannot Simplify\n\n";
      | Some (te, ty) ->
          state.lastcmd <- Command :: state.lastcmd;
          printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);

  }

| SHOWDEF DOT {

    let l = vmext state.def in
      
      printf "Definitions:\n";

      List.fold_left (

        fun tl hd ->
          let (name, (te, ty)) = hd in
            printf "%s := %s : %s\n"
              name
              (string_of_term te state.implicite)
              (string_of_term ty state.implicite);

      ) () l; printf "\n";
      state.lastcmd <- Command :: state.lastcmd;

  }

| SHOWCOERCION DOT {

    let l = state.coercion in
      
      printf "Coercions:\n";

      List.fold_left (

        fun tl hd ->
          let (te, ty, i) = hd in
            printf " %s: %s  [%d]\n"
              (string_of_term te state.implicite)
              (string_of_term ty state.implicite)
              i;

      ) () l; printf "\n";
      state.lastcmd <- Command :: state.lastcmd;

  }

| UNDO DOT {

    match state.lastcmd with
      | [] -> ();
      | hd :: tl -> (

          (
            match hd with
              | Definition l -> (
                  
                  List.fold_left (
                    
                    fun acc hd ->
                      
                      state.def <- VarMap.remove hd state.def;

                  ) () l;
                  
                )
                  
              | Command -> (
                  
                  ();
                  
                )

              | Coercion -> (

                  match state.coercion with
                    | [] -> ();
                    | hd :: tl ->

                        state.coercion <- tl;                  

                )              

              | Overload (s, t) -> (

                  try (

                    let l = VarMap.find s state.overload in
                      
                    let l = List.fold_left (

                      fun acc hd ->

                        let (te, ty) = hd in

                          if (term_eq te t state.def) then
                            acc
                          else
                            acc @ hd::[]

                    ) [] l in

                      state.overload <- VarMap.add s l (VarMap.remove s state.overload);

                  ) with

                    | _ -> ();

                )

              | Implicite s -> (

                  state.implicite <- VarMap.remove s state.implicite;

                )

              (* To complete *)

              | Lemma s -> (

                  state.def <- VarMap.remove s state.def;

                )

              | Tactic t -> (
                  
                  (
                    match state.pt with
                      | [] -> ();
                      | hd1 :: tl1 ->
                          state.pt <- tl1;
                  );
                  state.lastcmd <- tl;
                  (if (not (List.length state.pt = 0)) then             
                    print_subgoals_ctxt()
                    else 
                      ()
                  );                          
                                    
                )

          );

          state.lastcmd <- tl;

        )

  }

| QUIT DOT {
    exit 0;
  }


| LEMMA NAME DDOT term DOT {
  
    if (proof_mode()) then 
      
      printf "Error: actually already in proof mode\n\n"

    else (
      
      match check_term $4 (Some Type) state.def state.coercion state.oracles state.overload state.implicite true with
          
        | Some (te, ty) -> (
            
            state.subgoals <- 0;
            let subgoal_name = (concat "" ("subgoal"::(string_of_int state.subgoals):: [])) in 
              
              state.pt <- (([],subgoal_name, te)::[], ($2, Var subgoal_name, te)) :: [];          
              state.lastcmd <- Lemma $2 :: state.lastcmd;                            
              print_subgoals_ctxt()
              
          )
            
        | _ -> (
            
            printf "Error: invalide type !\n\n";
            
          )
            
    )
  }
    
| QED DOT {
    
    match state.pt with
      | [] -> printf "Error: Not in proof mode\n\n";
      | ([], (n, te,ty)) :: tl -> (
          match check_term te (Some ty) state.def state.coercion state.oracles state.overload state.implicite true                   
          with
            | Some (te, ty) -> (

                printf "proof complete.\n\n";
                state.pt <- [];
                state.subgoals <- 0;
                state.def <- VarMap.add n (Obj ((new axiom n ty) :> term pObj), ty) state.def;                  
                state.lastcmd <- (remove_all_tactic (state.lastcmd));
                
              )
            | _ ->
                printf "Error: proof does not type!\n\n";
                print_term te state.implicite; printf "\n\n";
            
        )
      | _ -> 
          printf "proof not complete!\n\n";
          print_subgoals_ctxt()
            
  }

| tactic DOT { 
    match state.pt with
      | [] -> printf "Error: Not in proof mode\n\n";
      | (lgh, lgg) :: tl -> (
          match (applytactics $1 1 lgh lgg  state.def state.coercion state.oracles state.overload state.implicite) with
            | None -> printf "Error: Cannot apply Tactic\n\n";
            | Some (lgh,lgg,n) ->
                printf "%d subgoals generated\n\n" n;
                state.pt <- (lgh,lgg) :: state.pt;
                state.lastcmd <- Tactic $1 :: state.lastcmd;                                     
 
        );
          print_subgoals_ctxt();
  
  }


;


  definition : 

| INDUCTIVE NAME listargs DDOT term DDOT EQ constructors DOT {

    let te = Ind ($2, $3, $5, list_proj2 $8) in 
      match check_term te None state.def state.coercion state.oracles state.overload state.implicite true with
        | None -> printf "Cannot Type Check\n\n";
        | Some (te, ty) ->

            let l = (VarSet.inter ($2 ++ (listvarterm2varset $8)) ((vmdomain state.def) +++ (vmdomain state.overload))) in
              if (not (VarSet.is_empty l)) then (
                
                printf "(%s) already defined\n\n" (string_of_listofvar (VarSet.elements l));
                  
              ) else (                
                
                printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);

                match te with
                  | Ind (n, la, ty2, lc) -> (

                      let def = VarMap.add $2 (te, (*(largs_foralls (var2cste ty ((vmdomain state.def) +++ (vmdomain state.overload))) la)*) ty) state.def in
                        
                      let (lc2, def,_) = 
                        List.fold_left (
                          
                          fun l hd ->
                            let (lcmd, def, n) = l in
                            let (cname, ctype) = hd in
                              (lcmd @ cname::[],
                                (VarMap.add 
                                  cname 
                                  (Cons (n,Cste $2), (largs_foralls (var2cste (List.nth lc n) ((vmdomain def) +++ (vmdomain state.overload))) la)) 
                                  def
                              ),n+1)
                                
                                
                        ) ($2::[], def, 0) $8 in
                        
                      let ipty = ind_princ_type n la ty2 lc def in
                      let ipte = ind_princ_pf n la ty2 lc def in
                      let (lc2, def) = (
                        match check_term ipte (Some ipty) def state.coercion state.oracles state.overload state.implicite true with
                          | None -> (lc2, def)
                          | Some (te, ty) ->
                              let name = concat "" (n :: "induc" :: []) in
                                (name :: lc2, VarMap.add name (fold te def, fold ty def) def)
                      ) in
                        
                        state.def <- def;
                        state.lastcmd <- Definition lc2 :: state.lastcmd;
                        
                        
                    )
                  | _ -> ();
           
              )

  }

| RECURSIVE NAME listargs LBRACK listnum RBRACK DDOT term DDOT EQ term DOT {

    
    let te = Phi ($2, $3, $8, $5, $11) in 
      match check_term te None state.def state.coercion state.oracles state.overload state.implicite true with
        | None -> printf "Cannot Type Check\n\n";
        | Some (te, ty) ->

            let l = (VarSet.inter (VarSet.singleton $2) ((vmdomain state.def) +++ (vmdomain state.overload))) in
              if (not (VarSet.is_empty l)) then (
                
                printf "(%s) already defined\n\n" (string_of_listofvar (VarSet.elements l));
                  
              ) else (                

                state.def <- VarMap.add $2 (te, ty) state.def;
                state.lastcmd <- Definition ($2::[]) :: state.lastcmd;
                printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);

              )

  }

| DEFINITION NAME DDOT term DDOT EQ term DOT {

      match check_term $7 (Some $4) state.def state.coercion state.oracles state.overload state.implicite true with
        | None -> printf "Cannot Type Check\n\n";
        | Some (te, ty) ->
            let l = (VarSet.inter (VarSet.singleton $2) ((vmdomain state.def) +++ (vmdomain state.overload))) in
              if (not (VarSet.is_empty l)) then (
                
                printf "(%s) already defined\n\n" (string_of_listofvar (VarSet.elements l));
                  
              ) else (                

                state.def <- VarMap.add $2 (te, ty) state.def;
                state.lastcmd <- Definition ($2::[]) :: state.lastcmd;
                printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);
            
              )

  }

| DEFINITION NAME DDOT EQ term DOT {

      match check_term $5 None state.def state.coercion state.oracles state.overload state.implicite true with
        | None -> printf "Cannot Type Check\n\n";
        | Some (te, ty) ->
            let l = (VarSet.inter (VarSet.singleton $2) ((vmdomain state.def) +++ (vmdomain state.overload))) in
              if (not (VarSet.is_empty l)) then (
                
                printf "(%s) already defined\n\n" (string_of_listofvar (VarSet.elements l));
                
              ) else (                

                state.def <- VarMap.add $2 (te, ty) state.def;
                state.lastcmd <- Definition ($2::[]) :: state.lastcmd;
                printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);
            
              )

  }

| DEFINITION NAME listargs DDOT term DDOT EQ term DOT {

      match check_term (largs_lambdas $8 $3) (Some (largs_foralls $5 $3)) state.def state.coercion state.oracles state.overload state.implicite true with
        | None -> printf "Cannot Type Check\n\n";
        | Some (te, ty) ->
            let l = (VarSet.inter (VarSet.singleton $2) ((vmdomain state.def) +++ (vmdomain state.overload))) in
              if (not (VarSet.is_empty l)) then (
                
                printf "(%s) already defined\n\n" (string_of_listofvar (VarSet.elements l));
                  
              ) else (                

                state.def <- VarMap.add $2 (te, ty) state.def;
                state.lastcmd <- Definition ($2::[]) :: state.lastcmd;
                printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);
            
              )

  }

| DEFINITION NAME listargs DDOT EQ term DOT {

      match check_term (largs_lambdas $6 $3) None state.def state.coercion state.oracles state.overload state.implicite true with
        | None -> printf "Cannot Type Check\n\n";
        | Some (te, ty) ->
            let l = (VarSet.inter (VarSet.singleton $2) ((vmdomain state.def) +++ (vmdomain state.overload))) in
              if (not (VarSet.is_empty l)) then (
                
                printf "(%s) already defined\n\n" (string_of_listofvar (VarSet.elements l));
                  
              ) else (                

                state.def <- VarMap.add $2 (te, ty) state.def;
                state.lastcmd <- Definition ($2::[]) :: state.lastcmd;
                printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);
            
              )

  }


| COERCION term LBRACK NUM RBRACK DOT {

    match check_term $2 None state.def state.coercion state.oracles state.overload state.implicite true with
      | None -> printf "Cannot Type Check\n\n";
      | Some (te, ty) ->
          
          state.coercion <- (te, ty, $4)::state.coercion;
          state.lastcmd <- Coercion :: state.lastcmd;
          printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);

  }


| OVERLOAD NAME WITH term LBRACK NUM RBRACK DOT {

    match check_term $4 None state.def state.coercion state.oracles state.overload state.implicite true with
      | None -> ();

      | Some (te, ty) -> 
          
          try (

            let l = VarMap.find $2 state.overload in

              state.overload <- VarMap.add $2 ((te,$6)::l) (VarMap.remove $2 state.overload);
              state.lastcmd <- Overload ($2, te) :: state.lastcmd;

          ) with
            | _ -> 
                state.overload <- VarMap.add $2 ((te,$6)::[]) state.overload;
                state.lastcmd <- Overload ($2, te) :: state.lastcmd;


  }

| IMPLICITE NAME LBRACK NUM RBRACK DOT {

    try (
      
      let _ = VarMap.find $2 state.implicite in
        
        ()
          
          
    ) with
      | _ -> 
          state.implicite <- VarMap.add $2 $4 state.implicite;
          state.lastcmd <- Implicite $2 :: state.lastcmd          


  }

;



 term:
| terma {

    $1
      
  }
| terma suiteterm {
    let te1 = $1 in
    let te2 = $2 in
      App (te1,te2)
  }
| term IMPL term {
    
    let te = $1 in
    let te2 = $3 in
    let x = "?H" in
      Forall (x, te, te2)

}
 
;

  terma:

| UNDERSCORE {

    Var "?T"

  }

| NAME { 

    try (
      
      let i = VarMap.find $1 state.implicite in

      let l = list_of_n_elt "?T" i in

        App (Cste $1, (list_name2list_var l))

    ) with
      | _ ->  Var $1

  }

| AROBAS NAME {

    Cste $2

  }

| LAMBDA LPAREN NAME DDOT term RPAREN DOT term {

    let ty = $5 in
    let te = $8 in
      Lambda ($3, ty, te)

  }

| LAMBDA NAME DOT term {

    let te = $4 in
    let x = "?T" in
      Lambda ($2, (Var x), te)

  }

| FORALL LPAREN NAME DDOT term RPAREN DOT term {

    let ty = $5 in
    let te = $8 in
      Forall ($3, ty, te)

  }

| FORALL NAME DOT term {

    let te = $4 in
    let x = "?T" in
      Forall ($2, (Var x), te)

  }

| MATCH term WITH destructors {
    Match ($2, $4, None)
  }

| MATCH term RETURN term WITH destructors {
    Match ($2, $6, Some $4)
  }

| CONS NUM OF term {
    Cons ($2, $4)
  }

| TYPE { Type }

| LPAREN term RPAREN { $2 }

/* Only for testing */


| PHI LPAREN NAME COMMA LPAREN context RPAREN COMMA term COMMA LBRACK listnum RBRACK RPAREN DOT term {
    Phi ($3, $6, $9, $12, $16)
  }

| IND LPAREN NAME COMMA LPAREN context RPAREN COMMA term RPAREN DOT listterm {
    Ind ($3, $6, $9, $12)
  }


;

  listterm:
| term { $1 :: [] }
| listterm COMMA listterm { $1 @ $3 }
| LPAREN listterm RPAREN { $2 }
| LPAREN RPAREN { [] }
;

  suiteterm:
| terma { $1::[] }
| terma suiteterm { $1 :: $2 }
;


  args:

| LPAREN NAME DDOT term RPAREN { ($2,$4) }
| NAME { 
        
    ($1, Var "?T")

  }
;

  context:
| args { $1::[] }
| args COMMA context { $1::$3 }
| { [] }
;

  listargs:

| args { $1::[] }
| args listargs { $1 :: $2}
| { [] }
;

 destructors:
| destructors destructors {$1 @ $2 }
| PAR term { $2::[] }
| { [] } 
;

 listnum:
| NUM { $1::[] }
| listnum COMMA listnum {$1 @ $3}
| { [] }
;

  constructors:
| constructors constructors { $1 @ $2 }
| PAR NAME DDOT term { ($2, $4)::[] }
| { [] }
;


 tactic:

| LPAREN tactic RPAREN {

    $2

  }

| tactic SCOMMA tactic {

    TacticSeq ($1, $3)

  }

| DO NUM tactic { Dotactic ($2, $3) }

| TRY tactic { Trytactic $2 }

| LBRACK listtactic RBRACK { Partactic $2 }

| LPAREN listtactic2 RPAREN { Ortactic $2 }

| REPEAT tactic { Repeat $2 }

| EQUIV term {

    Equiv $2
      
  }


| IDTAC { Idtac }

| FAIL { Fail }

| INTRO {

      Intro

  }

| INTRO NAME {

    IntroAs $2

  }

| INTROS {

    Repeat Intro

  }

| EXACT term {
    match (state.pt) with
      | [] -> printf "Error: Not in proof mode\n\n"; Fail
      | lg :: tl -> 
          Exact $2
  }

| APPLY term {
      (match (state.pt) with
        | [] -> printf "Error: Not in proof mode\n\n"; Fail
        | lg ::tl -> 
            Apply $2
      )              
  }

| GENERALIZE term {

    (match (state.pt) with
      | [] -> printf "Error: Not in proof mode\n\n"; Fail
      | (lg1, (lg21, lg22, lg23)) :: tl -> 
          let name = (fresh_name_list_name (term_var lg22) "H") in
            Generalize ($2, name)
    )      
  }

| GENERALIZE term AS NAME {

      (match (state.pt) with
        | [] -> printf "Error: Not in proof mode\n\n"; Fail
        | lg :: tl -> 
            Generalize ($2, $4)
      )
      
  }

| SIMPL {

    (
      match state.pt with
        | [] -> printf "Error: Not in proof mode\n\n"; Fail
        | lg :: tl ->
            Simpl
    )
  }

| SIMPL IN NAME {

    (match state.pt with
      | [] -> printf "Error: Not in proof mode\n\n"; Fail
      | lg :: tl ->
          SimplIn $3
    )
       
  }

| FACTORIZE suiteterm {
      
    (
      match state.pt with
        | [] -> printf "Error: Not in proof mode\n\n"; Fail
        | lg :: tl ->
            Factorize $2
    )
  }

| DESTRUCT term {

    match state.pt with
      | [] -> printf "Error: Not in proof mode\n\n"; Fail
      | lg :: tl ->
          Destruct $2
            
  }



;

 listtactic:
   
| tactic { $1::[] }

| listtactic PAR listtactic { $1 @ $3 }

;

 listtactic2:
   
| tactic { $1::[] }

| listtactic2 PAR PAR listtactic2 { $1 @ $4 }

;

